from twisted.python.failure import Failure
from twisted.web import server
from .base import BaseResource


class AdminResource(BaseResource):
    
    isLeaf = True
    
    def __init__(self, adminserver):
        self.adminserver = adminserver
        BaseResource.__init__(self)
        
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        if len(request.postpath) > 0:
            if request.postpath[0] == "clear_http_cache":
                d = self.adminserver.clearHTTPCache()
                d.addCallback(self._successResponse)
                d.addErrback(self._errorResponse)
                d.addCallback(self._immediateResponse, request)
                return server.NOT_DONE_YET
        return self._errorResponse(Failure(exc_value=Exception("Unknown request."))) 




from twisted.web import server
import simplejson
from twisted.internet import reactor
from ..evaluateboolean import evaluateBoolean
import urlparse
from .base import BaseResource


class ExecutionResource(BaseResource):
    
    def __init__(self, executionserver):    
        self.executionserver = executionserver
        BaseResource.__init__(self)
    
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        if len(request.postpath) > 0:
            if request.postpath[0] == "coordinate":
                reactor.callLater(5, self.executionserver.coordinate)
                return simplejson.dumps(True)  
            elif request.postpath[0] == "server":
                return simplejson.dumps(self.executionserver.getServerData())
            elif request.postpath[0] == "getpage" and "url" in request.args:
                #print "Recieving peering request for %s" % request.args["url"][0]
                kwargs = {}
                if "method" in request.args:
                    kwargs["method"] = request.args["method"][0]     
                if "postdata" in request.args: 
                    kwargs["postdata"] = urlparse.parse_qs(request.args["postdata"][0])
                if "headers" in request.args: 
                    kwargs["headers"] = urlparse.parse_qs(request.args["headers"][0])  
                if "cookies" in request.args: 
                    kwargs["cookies"] = urlparse.parse_qs(request.args["cookies"][0])           
                if "agent" in request.args:
                    kwargs["agent"] = request.args["agent"][0]
                if "timeout" in request.args:
                    kwargs["timeout"] = int(request.args["timeout"][0])
                if "followRedirect" in request.args:
                    kwargs["followRedirect"] = evaluateBoolean(request.args["followRedirect"][0])
                if "url_hash" in request.args: 
                    kwargs["url_hash"] = request.args["url_hash"][0] 
                if "cache" in request.args: 
                    kwargs["cache"] = int(request.args["cache"][0])
                if "prioritize" in request.args: 
                    kwargs["prioritize"] = evaluateBoolean(request.args["prioritize"][0])
                d = self.executionserver.pg.getPage(request.args["url"][0], **kwargs)              
                d.addCallback(self._getpageCallback, request)         
                d.addErrback(self._errorResponse) 
                d.addCallback(self._immediateResponse, request)    
                return server.NOT_DONE_YET
        message = "No such resource."
        request.setResponseCode(404, message)
        self._immediateResponse(simplejson.dumps({"error":message}), request)
    
    def _getpageCallback(self, data, request):
        request.setResponseCode(data["status"], data["message"])
        for header in data["headers"]:
            request.setHeader(header, data["headers"][header][0])
        return self._immediateResponse(data["response"], request)


from .base import BaseResource

class WorkerResource(BaseResource):
    
    def __init__(self, heapserver):    
        self.heapserver = heapserver
        BaseResource.__init__(self)
    
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        return "{}"


"""
Twisted API for Amazon's various web services.
"""

from .productadvertising import AmazonProductAdvertising
from .s3 import AmazonS3
from .sqs import AmazonSQS
from .sdb import AmazonSDB
from .sdb import sdb_now
from .sdb import sdb_now_add
from .sdb import sdb_parse_time
from .sdb import sdb_latitude
from .sdb import sdb_longitude



from twisted.web import server
from .base import BaseResource
from twisted.python.failure import Failure
from twisted.internet.defer import DeferredList

class InterfaceResource(BaseResource):
    
    isLeaf = True
    
    def __init__(self, interfaceserver):
        self.interfaceserver = interfaceserver
        BaseResource.__init__(self)
    
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        if len(request.postpath) > 0:
            if request.postpath[0] == "show_reservation":
                if "uuid" in request.args:
                    deferreds = []
                    for uuid in request.args["uuid"]:
                        deferreds.append(self.interfaceserver.showReservation(uuid))
                    d = DeferredList(deferreds, consumeErrors=True)
                    d.addCallback(self._showReservationCallback, request.args["uuid"])
                    d.addCallback(self._successResponse)
                    d.addErrback(self._errorResponse)
                    d.addCallback(self._immediateResponse, request)
                    return server.NOT_DONE_YET
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required.")))
            elif request.postpath[0] == "execute_reservation":
                if "uuid" in request.args:
                    d = self.interfaceserver.executeReservation(request.args["uuid"][0])
                    d.addCallback(self._successResponse)
                    d.addErrback(self._errorResponse)
                    d.addCallback(self._immediateResponse, request)
                    return server.NOT_DONE_YET
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required.")))
            elif request.postpath[0] == "delete_reservation":
                if "uuid" in request.args:
                    d = self.interfaceserver.deleteReservation(request.args["uuid"][0])
                    d.addCallback(self._successResponse)
                    d.addErrback(self._errorResponse)
                    d.addCallback(self._immediateResponse, request)
                    return server.NOT_DONE_YET
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required.")))

    def _showReservationCallback(self, data, uuids):
        print data[0][1]
        response = {}
        for i in range(0, len(uuids)):
            if data[i][0] == True:
                response[uuids[i]] = data[i][1]
            else:
                response[uuids[i]] = {"error":str(data[i][1].value)}
        return response




import specs
from twisted.internet.protocol import ClientCreator
from twisted.internet import reactor
from txamqp.client import TwistedDelegate
from txamqp.protocol import AMQClient
import txamqp.spec

def createClient(amqp_host, amqp_vhost, amqp_port=5672):
    amqp_spec = txamqp.spec.loadString(specs.v0_8)
    amqp_delegate = TwistedDelegate()
    client = ClientCreator(reactor, 
        AMQClient, 
        delegate=amqp_delegate,
        vhost=amqp_vhost,
        spec=amqp_spec).connectTCP(amqp_host, amqp_port)
    return client


import cPickle
import twisted.python.failure
import datetime
import dateutil.parser
import hashlib
import logging
import time
import copy
from twisted.internet.defer import maybeDeferred
from .requestqueuer import RequestQueuer
from .unicodeconverter import convertToUTF8, convertToUnicode
from .exceptions import StaleContentException

class ReportedFailure(twisted.python.failure.Failure):
    pass

# A UTC class.
class CoordinatedUniversalTime(datetime.tzinfo):
    
    ZERO = datetime.timedelta(0)
    
    def utcoffset(self, dt):
        return self.ZERO
        
    def tzname(self, dt):
        return "UTC"
        
    def dst(self, dt):
        return self.ZERO


UTC = CoordinatedUniversalTime()
LOGGER = logging.getLogger("main")


class PageGetter:
    
    def __init__(self, 
        s3, 
        aws_s3_http_cache_bucket,
        time_offset=0,
        rq=None):
        """
        Create an S3 based HTTP cache.

        **Arguments:**
         * *s3* -- S3 client object.
         * *aws_s3_http_cache_bucket* -- S3 bucket to use for the HTTP cache.

        **Keyword arguments:**
         * *rq* -- Request Queuer object. (Default ``None``)      

        """
        self.s3 = s3
        self.aws_s3_http_cache_bucket = aws_s3_http_cache_bucket
        self.time_offset = time_offset
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
    
    def clearCache(self):
        """
        Clear the S3 bucket containing the S3 cache.
        """
        d = self.s3.emptyBucket(self.aws_s3_http_cache_bucket)
        return d
        
    def getPage(self, 
            url, 
            method='GET', 
            postdata=None,
            headers=None, 
            agent="AWSpider", 
            timeout=60, 
            cookies=None, 
            follow_redirect=1, 
            prioritize=False,
            hash_url=None, 
            cache=0,
            content_sha1=None,
            confirm_cache_write=False):
        """
        Make a cached HTTP Request.

        **Arguments:**
         * *url* -- URL for the request.

        **Keyword arguments:**
         * *method* -- HTTP request method. (Default ``'GET'``)
         * *postdata* -- Dictionary of strings to post with the request. 
           (Default ``None``)
         * *headers* -- Dictionary of strings to send as request headers. 
           (Default ``None``)
         * *agent* -- User agent to send with request. (Default 
           ``'AWSpider'``)
         * *timeout* -- Request timeout, in seconds. (Default ``60``)
         * *cookies* -- Dictionary of strings to send as request cookies. 
           (Default ``None``).
         * *follow_redirect* -- Boolean switch to follow HTTP redirects. 
           (Default ``True``)
         * *prioritize* -- Move this request to the front of the request 
           queue. (Default ``False``)
         * *hash_url* -- URL string used to indicate a common resource.
           Example: "http://digg.com" and "http://www.digg.com" could both
           use hash_url, "http://digg.com" (Default ``None``)      
         * *cache* -- Cache mode. ``1``, immediately return contents of 
           cache if available. ``0``, check resource, return cache if not 
           stale. ``-1``, ignore cache. (Default ``0``)
         * *content_sha1* -- SHA-1 hash of content. If this matches the 
           hash of data returned by the resource, raises a 
           StaleContentException.  
         * *confirm_cache_write* -- Wait to confirm cache write before returning.       
        """       
        request_kwargs = {
            "method":method.upper(), 
            "postdata":postdata, 
            "headers":headers, 
            "agent":agent, 
            "timeout":timeout, 
            "cookies":cookies, 
            "follow_redirect":follow_redirect, 
            "prioritize":prioritize}
        cache = int(cache)
        if cache not in [-1,0,1]:
            raise Exception("Unknown caching mode.")
        if not isinstance(url, str):
            url = convertToUTF8(url)
        if hash_url is not None and not isinstance(hash_url, str):
            hash_url = convertToUTF8(hash_url)
        # Create request_hash to serve as a cache key from
        # either the URL or user-provided hash_url.
        if hash_url is None:
            request_hash = hashlib.sha1(cPickle.dumps([
                url, 
                headers, 
                agent, 
                cookies])).hexdigest()
        else:
            request_hash = hashlib.sha1(cPickle.dumps([
                hash_url, 
                headers, 
                agent, 
                cookies])).hexdigest()
        if request_kwargs["method"] != "GET":
            d = self.rq.getPage(url, **request_kwargs)
            d.addCallback(self._checkForStaleContent, content_sha1, request_hash)
            return d
        if cache == -1:
            # Cache mode -1. Bypass cache entirely.
            LOGGER.debug("Getting request %s for URL %s." % (request_hash, url))
            d = self.rq.getPage(url, **request_kwargs)
            d.addCallback(self._returnFreshData, 
                request_hash, 
                url,  
                confirm_cache_write)
            d.addErrback(self._requestWithNoCacheHeadersErrback,
                request_hash, 
                url, 
                confirm_cache_write,
                request_kwargs)
            d.addCallback(self._checkForStaleContent, content_sha1, request_hash)
            return d
        elif cache == 0:
            # Cache mode 0. Check cache, send cached headers, possibly use cached data.
            LOGGER.debug("Checking S3 Head object request %s for URL %s." % (request_hash, url))
            # Check if there is a cache entry, return headers.
            d = self.s3.headObject(self.aws_s3_http_cache_bucket, request_hash)
            d.addCallback(self._checkCacheHeaders, 
                request_hash,
                url,  
                request_kwargs,
                confirm_cache_write,
                content_sha1)
            d.addErrback(self._requestWithNoCacheHeaders, 
                request_hash, 
                url, 
                request_kwargs,
                confirm_cache_write)  
            d.addCallback(self._checkForStaleContent, content_sha1, request_hash)    
            return d
        elif cache == 1:
            # Cache mode 1. Use cache immediately, if possible.
            LOGGER.debug("Getting S3 object request %s for URL %s." % (request_hash, url))
            d = self.s3.getObject(self.aws_s3_http_cache_bucket, request_hash)
            d.addCallback(self._returnCachedData, request_hash)
            d.addErrback(self._requestWithNoCacheHeaders, 
                request_hash, 
                url, 
                request_kwargs,
                confirm_cache_write)    
            d.addCallback(self._checkForStaleContent, content_sha1, request_hash)    
            return d      
                  
    def _checkCacheHeaders(self, 
            data, 
            request_hash, 
            url, 
            request_kwargs,
            confirm_cache_write,
            content_sha1):
        LOGGER.debug("Got S3 Head object request %s for URL %s." % (request_hash, url))
        http_history = {}
        #if "content-length" in data["headers"] and int(data["headers"]["content-length"][0]) == 0:
        #    raise Exception("Zero Content length, do not use as cache.")
        if "content-sha1" in data["headers"]:
            http_history["content-sha1"] = data["headers"]["content-sha1"][0]
        # Filter?
        if "request-failures" in data["headers"]:
            http_history["request-failures"] = data["headers"]["request-failures"][0].split(",")
        if "content-changes" in data["headers"]:
            http_history["content-changes"] = data["headers"]["content-changes"][0].split(",")
        # If cached data is not stale, return it.
        if "cache-expires" in data["headers"]:
            expires = dateutil.parser.parse(data["headers"]["cache-expires"][0])
            now = datetime.datetime.now(UTC)
            if expires > now:
                if "content-sha1" in http_history and http_history["content-sha1"] == content_sha1:
                    LOGGER.debug("Raising StaleContentException (1) on %s" % request_hash)
                    raise StaleContentException()
                LOGGER.debug("Cached data %s for URL %s is not stale. Getting from S3." % (request_hash, url))
                d = self.s3.getObject(self.aws_s3_http_cache_bucket, request_hash)
                d.addCallback(self._returnCachedData, request_hash)
                d.addErrback(
                    self._requestWithNoCacheHeaders, 
                    request_hash, 
                    url,
                    request_kwargs, 
                    confirm_cache_write,
                    http_history=http_history)
                return d
        modified_request_kwargs = copy.deepcopy(request_kwargs)
        # At this point, cached data may or may not be stale.
        # If cached data has an etag header, include it in the request.
        if "cache-etag" in data["headers"]:
            modified_request_kwargs["etag"] = data["headers"]["cache-etag"][0]
        # If cached data has a last-modified header, include it in the request.
        if "cache-last-modified" in data["headers"]:
            modified_request_kwargs["last_modified"] = data["headers"]["cache-last-modified"][0]
        LOGGER.debug("Requesting %s for URL %s with etag and last-modified headers." % (request_hash, url))
        # Make the request. A callback means a 20x response. An errback 
        # could be a 30x response, indicating the cache is not stale.
        d = self.rq.getPage(url, **modified_request_kwargs)
        d.addCallback(
            self._returnFreshData, 
            request_hash,
            url, 
            confirm_cache_write,
            http_history=http_history)
        d.addErrback(
            self._handleRequestWithCacheHeadersError, 
            request_hash, 
            url, 
            request_kwargs, 
            confirm_cache_write,
            data,
            http_history,
            content_sha1)
        return d
        
    def _returnFreshData(self, 
            data, 
            request_hash, 
            url,  
            confirm_cache_write,
            http_history=None):
        LOGGER.debug("Got request %s for URL %s." % (request_hash, url))
        data["pagegetter-cache-hit"] = False
        data["content-sha1"] = hashlib.sha1(data["response"]).hexdigest()
        if http_history is not None and "content-sha1" in http_history:
            if http_history["content-sha1"] == data["content-sha1"]:
                return data
        d = maybeDeferred(self._storeData,
            data, 
            request_hash,  
            confirm_cache_write,
            http_history=http_history)
        d.addErrback(self._storeDataErrback, data, request_hash)
        return d

    def _requestWithNoCacheHeaders(self, 
            error, 
            request_hash, 
            url, 
            request_kwargs, 
            confirm_cache_write,
            http_history=None):
        try:
            error.raiseException()
        except StaleContentException, e:
            LOGGER.debug("Raising StaleContentException (2) on %s" % request_hash)
            raise StaleContentException()
        except Exception, e:
            pass
        # No header stored in the cache. Make the request.
        LOGGER.debug("Unable to find header for request %s on S3, fetching from %s." % (request_hash, url))
        d = self.rq.getPage(url, **request_kwargs)
        d.addCallback(
            self._returnFreshData, 
            request_hash, 
            url, 
            confirm_cache_write,
            http_history=http_history)
        d.addErrback(
            self._requestWithNoCacheHeadersErrback, 
            request_hash, 
            url, 
            confirm_cache_write,
            request_kwargs,
            http_history=http_history)
        return d        
    
    def _requestWithNoCacheHeadersErrback(self, 
            error,     
            request_hash, 
            url, 
            confirm_cache_write,
            request_kwargs,
            http_history=None):
        LOGGER.error(error.value.__dict__)
        LOGGER.error("Unable to get request %s for URL %s.\n%s" % (
            request_hash, 
            url, 
            error))
        if http_history is None:
            http_history = {} 
        if "request-failures" not in http_history:
            http_history["request-failures"] = [str(int(self.time_offset + time.time()))]
        else:
            http_history["request-failures"].append(str(int(self.time_offset + time.time())))
        http_history["request-failures"] = http_history["request-failures"][-3:]
        LOGGER.debug("Writing data for failed request %s to S3." % request_hash)
        headers = {}
        headers["request-failures"] = ",".join(http_history["request-failures"])
        d = self.s3.putObject(
            self.aws_s3_http_cache_bucket, 
            request_hash, 
            "", 
            content_type="text/plain", 
            headers=headers)
        if confirm_cache_write:
            d.addCallback(self._requestWithNoCacheHeadersErrbackCallback, error)
            return d       
        return error
        
    def _requestWithNoCacheHeadersErrbackCallback(self, data, error):
        return error
    
    def _handleRequestWithCacheHeadersError(self, 
            error, 
            request_hash, 
            url, 
            request_kwargs,  
            confirm_cache_write,
            data,
            http_history,
            content_sha1):
        if error.value.status == "304":
            if "content-sha1" in http_history and http_history["content-sha1"] == content_sha1:
                LOGGER.debug("Raising StaleContentException (3) on %s" % request_hash)
                raise StaleContentException()
            LOGGER.debug("Request %s for URL %s hasn't been modified since it was last downloaded. Getting data from S3." % (request_hash, url))
            d = self.s3.getObject(self.aws_s3_http_cache_bucket, request_hash)
            d.addCallback(self._returnCachedData, request_hash)
            d.addErrback(
                self._requestWithNoCacheHeaders, 
                request_hash, 
                url, 
                request_kwargs, 
                confirm_cache_write,
                http_history=http_history)
            return d
        else:
            if http_history is None:
                http_history = {} 
            if "request-failures" not in http_history:
                http_history["request-failures"] = [str(int(self.time_offset + time.time()))]
            else:
                http_history["request-failures"].append(str(int(self.time_offset + time.time())))
            http_history["request-failures"] = http_history["request-failures"][-3:]
            LOGGER.debug("Writing data for failed request %s to S3. %s" % (request_hash, error))
            headers = {}
            for key in data["headers"]:
                headers[key] = data["headers"][key][0]
            headers["request-failures"] = ",".join(http_history["request-failures"])
            d = self.s3.putObject(
                self.aws_s3_http_cache_bucket, 
                request_hash, 
                data["response"], 
                content_type=data["headers"]["content-type"][0], 
                headers=headers)
            if confirm_cache_write:
                d.addCallback(self._handleRequestWithCacheHeadersErrorCallback, error)
                return d
            return ReportedFailure(error)
            
    def _handleRequestWithCacheHeadersErrorCallback(self, data, error):
        return ReportedFailure(error)
        
    def _returnCachedData(self, data, request_hash):
        LOGGER.debug("Got request %s from S3." % (request_hash))
        data["pagegetter-cache-hit"] = True
        data["status"] = 304
        data["message"] = "Not Modified"
        if "content-sha1" in data["headers"]:
            data["content-sha1"] = data["headers"]["content-sha1"][0]
            del data["headers"]["content-sha1"]
        else:
            data["content-sha1"] = hashlib.sha1(data["response"]).hexdigest()
        if "cache-expires" in data["headers"]:
            data["headers"]["expires"] = data["headers"]["cache-expires"]
            del data["headers"]["cache-expires"]
        if "cache-etag" in data["headers"]:
            data["headers"]["etag"] = data["headers"]["cache-etag"]
            del data["headers"]["cache-etag"]
        if "cache-last-modified" in data["headers"]:
            data["headers"]["last-modified"] = data["headers"]["cache-last-modified"]
            del data["headers"]["cache-last-modified"]
        return data
            
    def _storeData(self, 
            data, 
            request_hash,  
            confirm_cache_write,
            http_history=None):
        if len(data["response"]) == 0:
            return self._storeDataErrback(Failure(exc_value=Exception("Response data is of length 0")), response_data, request_hash)
        #data["content-sha1"] = hashlib.sha1(data["response"]).hexdigest()
        if http_history is None:
            http_history = {} 
        if "content-sha1" not in http_history:
            http_history["content-sha1"] = data["content-sha1"]
        if "content-changes" not in http_history:
            http_history["content-changes"] = []
        if data["content-sha1"] != http_history["content-sha1"]:
            http_history["content-changes"].append(str(int(self.time_offset + time.time())))
        http_history["content-changes"] = http_history["content-changes"][-10:]
        LOGGER.debug("Writing data for request %s to S3." % request_hash)
        headers = {}
        http_history["content-changes"] = filter(lambda x:len(x) > 0, http_history["content-changes"])
        headers["content-changes"] = ",".join(http_history["content-changes"])
        headers["content-sha1"] = data["content-sha1"]
        if "cache-control" in data["headers"]: 
            if "no-cache" in data["headers"]["cache-control"][0]:
                return data
        if "expires" in data["headers"]:
            headers["cache-expires"] = data["headers"]["expires"][0]
        if "etag" in data["headers"]:
            headers["cache-etag"] = data["headers"]["etag"][0]
        if "last-modified" in data["headers"]:
            headers["cache-last-modified"] = data["headers"]["last-modified"][0]
        if "content-type" in data["headers"]:
            content_type = data["headers"]["content-type"][0]
        d = self.s3.putObject(
            self.aws_s3_http_cache_bucket, 
            request_hash, 
            data["response"], 
            content_type=content_type, 
            headers=headers)
        if confirm_cache_write:
            d.addCallback(self._storeDataCallback, data)
            d.addErrback(self._storeDataErrback, data, request_hash)
            return d
        return data
        
    def _storeDataCallback(self, data, response_data):
        return response_data
    
    def _storeDataErrback(self, error, response_data, request_hash):
        LOGGER.error("Error storing data for %s" % (request_hash))
        return response_data

    def _checkForStaleContent(self, data, content_sha1, request_hash):
        if "content-sha1" not in data:
            data["content-sha1"] = hashlib.sha1(data["response"]).hexdigest()
        if content_sha1 == data["content-sha1"]:
            LOGGER.debug("Raising StaleContentException (4) on %s" % request_hash)
            raise StaleContentException(content_sha1)
        else:
            return data
            




import base64
import hmac
import hashlib
import urllib
import xml.etree.cElementTree as ET
from datetime import datetime
import time
import dateutil.parser
import logging
from twisted.internet.defer import DeferredList
from ..requestqueuer import RequestQueuer
from .lib import etree_to_dict, safe_quote_tuple


LOGGER = logging.getLogger("main")


SDB_NAMESPACE = "{http://sdb.amazonaws.com/doc/2009-04-15/}"

def base10toN(num,n):
    """Change a  to a base-n number.
    Up to base-36 is supported without special notation."""
    num_rep={10:'a',
         11:'b',
         12:'c',
         13:'d',
         14:'e',
         15:'f',
         16:'g',
         17:'h',
         18:'i',
         19:'j',
         20:'k',
         21:'l',
         22:'m',
         23:'n',
         24:'o',
         25:'p',
         26:'q',
         27:'r',
         28:'s',
         29:'t',
         30:'u',
         31:'v',
         32:'w',
         33:'x',
         34:'y',
         35:'z'}
    new_num_string=''
    current=num
    while current!=0:
        remainder=current%n
        if 36>remainder>9:
            remainder_string=num_rep[remainder]
        elif remainder>=36:
            remainder_string='('+str(remainder)+')'
        else:
            remainder_string=str(remainder)
        new_num_string=remainder_string+new_num_string
        current=current/n
    return new_num_string


def base10to36(i):
    return base10toN(i, 36)
        

def base36to10(s):
    return int(s, 36)


def sdb_now(offset=0):
    """Return an 11 character, zero padded string with the current Unixtime.
    
    **Keyword arguments:**
     * *offset* -- Offset in seconds. (Default 0)
    """
    return str(int(offset + time.time())).zfill(11)


def sdb_now_add(seconds, offset=0):
    """Return an 11 character, zero padded string with the current Unixtime
    plus an integer.
   
    **Arguments:**
     * *seconds* -- Seconds to add to the current time.
   
    **Keyword arguments:**
     * *offset* -- Offset in seconds. (Default 0)
    """
    return str(int(offset + time.time() + seconds)).zfill(11)


def sdb_parse_time(date_string, offset=0):
    """Parse a date string, then return an 11 character, zero padded
    string with the current Unixtime plus an integer.
   
    **Arguments:**
     * *date_string* -- Date string
   
    **Keyword arguments:**
     * *offset* -- Offset in seconds. (Default 0)
    """
    parsed_time = time.mktime(dateutil.parser.parse(date_string).timetuple())
    return str(int(offset + parsed_time)).zfill(11)


def sdb_latitude(latitude):
    """Return an 8 character, zero padded string version of the
    latitude parameter.
   
    **Arguments:**
     * *latitude* -- Latitude.
    """
    adjusted = (90 + float(latitude)) * 100000
    return str(int(adjusted)).zfill(8)


def sdb_longitude(longitude):
    """Return an 8 character, zero padded string version of the
    longitude parameter.
   
    **Arguments:**
     * *longitude* -- Longitude.
    """
    adjusted = (180 + float(longitude)) * 100000
    return str(int(adjusted)).zfill(8)

class AmazonSDB:
   
    """
    Amazon Simple Database API.
    """
   
    host = "sdb.amazonaws.com"
    box_usage = 0.0
   
    def __init__(self, aws_access_key_id, aws_secret_access_key, rq=None):
        """
        **Arguments:**
         * *aws_access_key_id* -- Amazon AWS access key ID
         * *aws_secret_access_key* -- Amazon AWS secret access key
       
        **Keyword arguments:**
         * *rq* -- Optional RequestQueuer object.
        """
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.rq.setHostMaxRequestsPerSecond(self.host, 0)
        self.rq.setHostMaxSimultaneousRequests(self.host, 0)

    def copyDomain(self, source_domain, destination_domain):
        """
        Copy all elements of a source domain to a destination domain.
       
        **Arguments:**
         * *source_domain* -- Source domain name
         * *destination_domain* -- Destination domain name
        """
        d = self.checkAndCreateDomain(destination_domain)
        d.addCallback(self._copyDomainCallback, source_domain, 
            destination_domain)
        return d
    
    def _copyDomainCallback(self, data, source_domain, destination_domain):
        return self._copyDomainCallback2(source_domain, destination_domain)
    
    def _copyDomainCallback2(self, source_domain, destination_domain, 
        next_token=None, total_box_usage=0):
        parameters = {}
        parameters["Action"] = "Select"
        parameters["SelectExpression"] = "SELECT * FROM `%s`" % source_domain
        if next_token is not None:
            parameters["NextToken"] = next_token
        d = self._request(parameters)
        d.addCallback(self._copyDomainCallback3, 
                      source_domain=source_domain,
                      destination_domain=destination_domain,
                      total_box_usage=total_box_usage)
        d.addErrback(self._genericErrback)
        return d

    def _copyDomainCallback3(self, data, source_domain, destination_domain,
        total_box_usage=0):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        total_box_usage += box_usage
        next_token_element = xml.find(".//%sNextToken" % SDB_NAMESPACE)
        if next_token_element is not None:
            next_token = next_token_element.text
        else:
            next_token = None
        items = xml.findall(".//%sItem" % SDB_NAMESPACE)
        results = {}
        for item in items:
            key = item.find("./%sName" % SDB_NAMESPACE).text
            attributes = item.findall("%sAttribute" % SDB_NAMESPACE)
            attribute_dict = {}
            for attribute in attributes:
                attr_name = attribute.find("./%sName" % SDB_NAMESPACE).text
                attr_value = attribute.find("./%sValue" % SDB_NAMESPACE).text
                if attr_name in attribute_dict:  
                    attribute_dict[attr_name].append(attr_value)
                else:
                    attribute_dict[attr_name] = [attr_value]
            results[key] = attribute_dict
        deferreds = []
        for key in results:
            d = self.putAttributes(destination_domain, key, results[key])
            d.addErrback(self._copyPutAttributesErrback, destination_domain, key, results[key])
            deferreds.append(d)
        d = DeferredList(deferreds, consumeErrors=True)
        d.addCallback(self._copyDomainCallback4, source_domain,
            destination_domain, next_token=next_token, total_box_usage=total_box_usage)
        return d
        
    def _copyDomainCallback4(self, data, source_domain, destination_domain,
        next_token=None, total_box_usage=0):
        for row in data:
            if row[0] == False:
                raise row[1]
        if next_token is not None:
            return self._copyDomainCallback2(
                source_domain=source_domain,
                destination_domain=destination_domain,
                next_token=next_token,
                total_box_usage=total_box_usage)
        LOGGER.debug("""CopyDomain:\n%s -> %s\nBox usage: %s""" % (
            source_domain,
            destination_domain,
            total_box_usage))
        return True
    
    def _copyPutAttributesErrback(self, error, destination_domain, key, attributes, count=0):
        if count < 3:
            d = self.putAttributes(destination_domain, key, attributes)
            d.addErrback(self._copyPutAttributesErrback, destination_domain, key, attributes, count=count + 1)
            return d
        return error
            
    def checkAndCreateDomain(self, domain): 
        """
        Check for a SimpleDB domain's existence. If it does not exist, 
        create it.
       
        **Arguments:**
         * *domain* -- Domain name
        """   
        d = self.domainMetadata(domain)
        d.addErrback(self._checkAndCreateDomainErrback, domain)  
        return d

    def _checkAndCreateDomainErrback(self, error, domain):
        if hasattr(error, "value") and hasattr(error.value, "status"):
            if int(error.value.status) == 400:  
                d = self.createDomain(domain)
                d.addErrback(self._checkAndCreateDomainErrback2, domain)
                return d
        message = "Could not find or create domain '%s'." % domain
        raise Exception(message)

    def _checkAndCreateDomainErrback2(self, error, domain):
        message =  "Could not create domain '%s'" % domain
        raise Exception(message)

    def createDomain(self, domain):
        """
        Create a SimpleDB domain.
       
        **Arguments:**
         * *domain* -- Domain name
        """
        parameters = {
            "Action":"CreateDomain",
            "DomainName":domain
        }
        d = self._request(parameters)
        d.addCallback(self._createDomainCallback, domain)
        d.addErrback(self._genericErrback)   
        return d

    def _createDomainCallback(self, data, domain):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("Created SimpleDB domain '%s'. Box usage: %s" % (domain,
            box_usage))
        return True
   
    def deleteDomain(self, domain):
        """
        Delete a SimpleDB domain.
       
        **Arguments:**
         * *domain* -- Domain name
        """
        parameters = {}
        parameters["Action"] = "DeleteDomain"
        parameters["DomainName"] = domain
        d = self._request(parameters)
        d.addCallback(self._deleteDomainCallback, domain)
        d.addErrback(self._genericErrback)
        return d

    def _deleteDomainCallback(self, data, domain):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("Deleted SimpleDB domain '%s'. Box usage: %s" % (domain, 
            box_usage))
        return True   

    def listDomains(self):
        """
        List SimpleDB domains associated with an account.
        """
        return self._listDomains()
   
    def _listDomains(self, 
                     next_token=None, 
                     previous_results=None, 
                     total_box_usage=0):
        parameters = {}
        parameters["Action"] = "ListDomains"
        if next_token is not None:
            parameters["NextToken"] = next_token
        d = self._request(parameters)
        d.addCallback(self._listDomainsCallback, 
                      previous_results=previous_results,
                      total_box_usage=total_box_usage)
        d.addErrback(self._genericErrback)   
        return d

    def _listDomainsCallback(self, 
                             data, 
                             previous_results=None, 
                             total_box_usage=0):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        total_box_usage += box_usage
        xml_response = etree_to_dict(xml, namespace=SDB_NAMESPACE)
        if "DomainName" in xml_response["ListDomainsResult"][0]:
            results = xml_response["ListDomainsResult"][0]["DomainName"]
        else:
            results = []
        if previous_results is not None:
            results.extend(previous_results)
        if "NextToken" in xml_response["ListDomainsResult"]:
            next_token = xml_response["ListDomainsResult"][0]["NextToken"][0]
            return self._listDomains(next_token=next_token,
                                     previous_results=results,
                                     total_box_usage=total_box_usage)
        LOGGER.debug("Listed domains. Box usage: %s" % total_box_usage)
        return results

    def domainMetadata(self, domain):
        """
        Return meta-information about a domain.
       
        **Arguments:**
         * *domain* -- Domain name
        """
        parameters = {}
        parameters["Action"] = "DomainMetadata"
        parameters["DomainName"] = domain
        d = self._request(parameters)
        d.addCallback(self._domainMetadataCallback, domain)
        d.addErrback(self._genericErrback)   
        return d
       
    def _domainMetadataCallback(self, data, domain):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("Got SimpleDB domain '%s' metadata. Box usage: %s" % (
            domain,
            box_usage))
        xml_response = etree_to_dict(xml, namespace=SDB_NAMESPACE)
        return xml_response["DomainMetadataResult"][0]
   
    def batchPutAttributes(self, domain, attributes_by_item_name, 
            replace_by_item_name=None):
        """
        Batch put attributes into domain.
       
        **Arguments:**
         * *domain* -- Domain name
         * *attributes_by_item_name* -- Dictionary of dictionaries. First 
           level keys are the item name, value is dictionary of key/value 
           pairs. Example: ``{"item_name":{"attribute_name":"value"}}``
         
         **Keyword arguments:**
         * *replace_by_item_name* -- Dictionary of lists. First level keys
           are the item names, value is a list of of attributes that should
           be overwritten. ``{"item_name":["attribute_name"]}`` (Default 
           empty dictionary)
        """   
        if replace_by_item_name is None:
            replace_by_item_name = {}
        if len(attributes_by_item_name) > 25:
            raise Exception("Too many items in batchPutAttributes. Up to 25 items per call allowed.")
        for item_name in replace_by_item_name:            
            if not isinstance(replace_by_item_name[item_name], list):
                raise Exception("Replace argument '%s' must be a list." % item_name)
        for item_name in attributes_by_item_name:
            if not isinstance(attributes_by_item_name[item_name], dict):
                raise Exception("Attributes argument '%s' must be a dictionary." % item_name)
        parameters = {}
        parameters["Action"] = "BatchPutAttributes"
        parameters["DomainName"] = domain
        i = 0
        for item_name in attributes_by_item_name:
            parameters["Item.%s.ItemName" % i] = item_name
            attributes_list = []
            for attribute in attributes_by_item_name[item_name].items():
                # If the attribute is a list, split into multiple attributes.
                if isinstance(attribute[1], list):
                    for value in attribute[1]:
                        attributes_list.append((attribute[0], value))
                else:
                    attributes_list.append(attribute)
            j = 0
            for attribute in attributes_list:
                parameters["Item.%s.Attribute.%s.Name" % (i,j)] = attribute[0]
                parameters["Item.%s.Attribute.%s.Value" % (i,j)] = attribute[1]
                if item_name in replace_by_item_name:
                    if attribute[0] in replace_by_item_name[item_name]:
                        parameters["Item.%s.Attribute.%s.Replace" % (i,j)] = "true"
                j += 1
            i += 1
        d = self._request(parameters)
        d.addCallback(
            self._batchPutAttributesCallback, 
            domain, 
            attributes_by_item_name)
        d.addErrback(self._genericErrback)
        return d
    
    def _batchPutAttributesCallback(self, 
            data, 
            domain, 
            attributes_by_item_name):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("""Batch put attributes %s in SimpleDB domain '%s'. Box usage: %s""" % (
            attributes_by_item_name,
            domain,
            box_usage))
        return True
        
    def putAttributes(self, domain, item_name, attributes, replace=None):
        """
        Put attributes into domain at item_name.
       
        **Arguments:**
         * *domain* -- Domain name
         * *item_name* -- Item name
         * *attributes* -- Dictionary of attributes
       
        **Keyword arguments:**
         * *replace* -- List of attributes that should be overwritten
           (Default empty list)
        """
        if replace is None:
            replace = []
        if not isinstance(replace, list):
            raise Exception("Replace argument must be a list.")
        if not isinstance(attributes, dict):
            raise Exception("Attributes argument must be a dictionary.")
        parameters = {}
        parameters["Action"] = "PutAttributes"
        parameters["DomainName"] = domain
        parameters["ItemName"] = item_name
        attributes_list = []
        for attribute in attributes.items():
            # If the attribute is a list, split into multiple attributes.
            if isinstance(attribute[1], list):
                for value in attribute[1]:
                    attributes_list.append((attribute[0], value))
            else:
                attributes_list.append(attribute)
        i = 0
        for attribute in attributes_list:
            parameters["Attribute.%s.Name" % i] = attribute[0]
            parameters["Attribute.%s.Value" % i] = attribute[1]
            if attribute[0] in replace:
                parameters["Attribute.%s.Replace" % i] = "true"
            i += 1
        d = self._request(parameters)
        d.addCallback(self._putAttributesCallback, domain, item_name, attributes)
        d.addErrback(self._genericErrback)
        return d
       
    def _putAttributesCallback(self, data, domain, item_name, attributes):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("""Put attributes %s on '%s' in SimpleDB domain '%s'. Box usage: %s""" % (
            attributes,
            item_name,
            domain,
            box_usage))
        return True
   
    def getAttributes(self, domain, item_name, attribute_name=None):
        """
        Get one or all attributes from domain at item_name.
       
        **Arguments:**
         * *domain* -- Domain name
         * *item_name* -- Item name
       
        **Keyword arguments:**
         * *attribute_name* -- Name of specific attribute to get (Default None)
        """
        parameters = {}
        parameters["Action"] = "GetAttributes"
        parameters["DomainName"] = domain
        parameters["ItemName"] = item_name
        if attribute_name is not None:
            parameters["AttributeName"] = attribute_name
        d = self._request(parameters)
        d.addCallback(self._getAttributesCallback, domain, item_name)
        d.addErrback(self._genericErrback)
        return d
       
    def _getAttributesCallback(self, data, domain, item_name):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("""Got attributes from '%s' in SimpleDB domain '%s'. Box usage: %s""" % (
            item_name,
            domain,
            box_usage))
        xml_response = etree_to_dict(xml, namespace=SDB_NAMESPACE)
        attributes = {}
        if xml_response["GetAttributesResult"][0] is None:
            raise Exception("Item does not exist.")
        for attribute in xml_response["GetAttributesResult"][0]['Attribute']:
            if attribute["Name"][0] not in attributes:
                attributes[attribute["Name"][0]] = []
            attributes[attribute["Name"][0]].extend(attribute["Value"])
        return attributes

    def delete(self, domain, item_name):
        """
        Delete all attributes from domain at item_name.
       
        **Arguments:**
         * *domain* -- Domain name
         * *item_name* -- Item name
        """
        return self.deleteAttributes(domain, item_name)

    def deleteAttributes(self, domain, item_name, attributes=None):
        """
        Delete one or all attributes from domain at item_name.
       
        **Arguments:**
         * *domain* -- Domain name
         * *item_name* -- Item name
       
        **Keyword arguments:**
         * *attributes* -- List of attribute names, or dictionary of
           attribute name / value pairs. (Default empty dict)
        """       
        if attributes is None:
            attributes = {}
        if not isinstance(attributes, dict) and \
           not isinstance(attributes, list):
            message = "Attributes parameter must be a dictionary or a list."
            raise Exception(message)
        parameters = {}
        parameters["Action"] = "DeleteAttributes"
        parameters["DomainName"] = domain
        parameters["ItemName"] = item_name
        if isinstance(attributes, dict):
            attr_count = 1
            for key in attributes:
                parameters["Attribute.%s.Name" % attr_count] = key
                parameters["Attribute.%s.Value" % attr_count] = attributes[key]
                attr_count += 1
        if isinstance(attributes, list):
            attr_count = 0
            for key in attributes:
                parameters["Attribute.%s.Name" % attr_count] = key
                attr_count += 1
        d = self._request(parameters)
        d.addCallback(self._deleteAttributesCallback, domain, item_name)
        d.addErrback(self._genericErrback)   
        return d

    def _deleteAttributesCallback(self, data, domain, item_name):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        LOGGER.debug("""Deleted attributes from '%s' in SimpleDB domain '%s'. Box usage: %s""" % (
            item_name,
            domain,
            box_usage))
        return True
   
    def select(self, select_expression, max_results=0):
        """
        Run a select query
       
        **Arguments:**
         * *select_expression* -- Select expression
        """  
        if "count(" in select_expression.lower():
            return self._selectCount(select_expression)
        return self._select(select_expression, max_results=max_results)
      
    def _selectCount(self, select_expression, next_token=None, 
            previous_count=0,
            total_box_usage=0):
        parameters = {}
        parameters["Action"] = "Select"
        parameters["SelectExpression"] = select_expression
        if next_token is not None:
            parameters["NextToken"] = next_token
        d = self._request(parameters)
        d.addCallback(self._selectCountCallback, 
                      select_expression=select_expression,
                      previous_count=previous_count,
                      total_box_usage=total_box_usage)
        d.addErrback(self._genericErrback)   
        return d

    def _selectCountCallback(self, data, select_expression=None, 
            previous_count=0,
            total_box_usage=0):
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        total_box_usage += box_usage
        next_token_element = xml.find(".//%sNextToken" % SDB_NAMESPACE)
        if next_token_element is not None:
            next_token = next_token_element.text
        else:
            next_token = None
        count = previous_count + int(xml.find(".//%sValue" % SDB_NAMESPACE).text)
        if next_token is not None:
            return self._selectCount(select_expression, next_token=next_token,
                previous_count=count,
                total_box_usage=total_box_usage)
        LOGGER.debug("""Select:\n'%s'\nBox usage: %s""" % (
            select_expression,
            total_box_usage))
        return count
        
    def _select(self, select_expression, next_token=None, 
            previous_results=None,
            total_box_usage=0,
            max_results=0):
        parameters = {}
        parameters["Action"] = "Select"
        parameters["SelectExpression"] = select_expression
        if next_token is not None:
            parameters["NextToken"] = next_token
        d = self._request(parameters)
        d.addCallback(self._selectCallback, 
                      select_expression=select_expression,
                      previous_results=previous_results,
                      total_box_usage=total_box_usage,
                      max_results=max_results)
        d.addErrback(self._genericErrback)   
        return d

    def _selectCallback(self, data, select_expression=None, 
            previous_results=None,
            total_box_usage=0,
            max_results=0):
        if previous_results is not None:
            results = previous_results
        else:
            results = {}
        xml = ET.fromstring(data["response"])
        box_usage = float(xml.find(".//%sBoxUsage" % SDB_NAMESPACE).text)
        self.box_usage += box_usage
        total_box_usage += box_usage
        next_token_element = xml.find(".//%sNextToken" % SDB_NAMESPACE)
        if next_token_element is not None:
            next_token = next_token_element.text
        else:
            next_token = None
        items = xml.findall(".//%sItem" % SDB_NAMESPACE)
        for item in items:
            key = item.find("./%sName" % SDB_NAMESPACE).text
            attributes = item.findall("%sAttribute" % SDB_NAMESPACE)
            attribute_dict = {}
            for attribute in attributes:
                attr_name = attribute.find("./%sName" % SDB_NAMESPACE).text
                attr_value = attribute.find("./%sValue" % SDB_NAMESPACE).text
                if attr_name in attribute_dict:  
                    attribute_dict[attr_name].append(attr_value)
                else:
                    attribute_dict[attr_name] = [attr_value]
            results[key] = attribute_dict
        if next_token is not None:
            if max_results == 0 or len(results) < max_results:
                return self._select(select_expression, next_token=next_token,
                                    previous_results=results,
                                    total_box_usage=total_box_usage,
                                    max_results=max_results)
        LOGGER.debug("""Select:\n'%s'\nBox usage: %s""" % (
            select_expression,
            total_box_usage))
        return results       

    def _request(self, parameters):
        """
        Add authentication parameters and make request to Amazon.
       
        **Arguments:**
         * *parameters* -- Key value pairs of parameters
        """
        parameters = self._getAuthorization("GET", parameters)
        query_string = urllib.urlencode(parameters)       
        url = "https://%s/?%s" % (self.host, query_string)
        if len(url) > 4096:
            del parameters['Signature']
            parameters = self._getAuthorization("POST", parameters)
            query_string = urllib.urlencode(parameters)       
            url = "https://%s" % (self.host)
            d = self.rq.getPage(url, method="POST", postdata=query_string)
            return d
        else:
            d = self.rq.getPage(url, method="GET")
            return d
         
    def _canonicalize(self, parameters):
        """
        Canonicalize parameters for use with AWS Authorization.
       
        **Arguments:**
         * *parameters* -- Key value pairs of parameters
       
        **Returns:**
         * A safe-quoted string representation of the parameters.
        """
        parameters = parameters.items()
        parameters.sort(lambda x, y:cmp(x[0], y[0]))
        return "&".join([safe_quote_tuple(x) for x in parameters])
       
    def _getAuthorization(self, method, parameters):
        """
        Create authentication parameters.
       
        **Arguments:**
         * *method* -- HTTP method of the request
         * *parameters* -- Key value pairs of parameters
       
        **Returns:**
         * A dictionary of authorization parameters
        """
        signature_parameters = {
            "AWSAccessKeyId":self.aws_access_key_id,
            "SignatureVersion":"2",
            "SignatureMethod":"HmacSHA256",
            'Timestamp':datetime.utcnow().isoformat()[0:19]+"+00:00",
            "AWSAccessKeyId":self.aws_access_key_id,
            "Version":"2009-04-15"
        }
        signature_parameters.update(parameters)
        query_string = self._canonicalize(signature_parameters)
        string_to_sign = "%(method)s\n%(host)s\n%(resource)s\n%(qs)s" % {
            "method":method,
            "host":self.host.lower(),
            "resource":"/",
            "qs":query_string,
        }
        args = [self.aws_secret_access_key, string_to_sign, hashlib.sha256]
        signature = base64.encodestring(hmac.new(*args).digest()).strip()
        signature_parameters.update({'Signature': signature})
        return signature_parameters

    def _genericErrback(self, error):
        if hasattr(error, "value"):
            if hasattr(error.value, "response"):
                xml = ET.XML(error.value.response)
                try:
                    LOGGER.debug(xml.find(".//Message").text)
                except Exception, e:
                    pass
        return error
        


import urllib
import re

__all__ = ["safe_quote_tuple", "etree_to_dict", "return_true"]

camelcase_to_underscore = lambda s: re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', s).lower().strip('_')

def safe_quote_tuple(tuple_):
    """Convert a 2-tuple to a string for use with AWS"""
    key = urllib.quote(str(tuple_[0]), '-_.~')
    value = urllib.quote(str(tuple_[1]), '-_.~')
    return "%s=%s" % (key, value)


def etree_to_dict(etree, namespace=None, tag_list=True, convert_camelcase=False):
    """
    Convert an etree to a dict.
   
    **Keyword arguments:**
     * *namespace* -- XML Namespace to be removed from tag names (Default None)
    """
    children = etree.getchildren()
    if len(children) == 0:
        return etree.text
    children_dict = {}
    for element in children:
        tag = element.tag
        if namespace is not None:
            tag = tag.replace(namespace, "")
        if convert_camelcase:
            tag = camelcase_to_underscore(tag)
        element_dict = etree_to_dict(element, namespace=namespace, tag_list=tag_list, convert_camelcase=convert_camelcase)
        if tag in children_dict:
            if not isinstance(children_dict[tag], list):
                children_dict[tag] = [children_dict[tag]]
            children_dict[tag].append(element_dict)
        else:
            if tag_list:    
                children_dict[tag] = [element_dict]
            else:
                children_dict[tag] = element_dict
    return children_dict


def return_true(data):
    return True



from twisted.internet.defer import Deferred, DeferredList, maybeDeferred
from twisted.web.resource import Resource
from twisted.internet import reactor
from twisted.internet import task
from twisted.web import server
from .base import BaseServer, LOGGER, PRETTYPRINTER
from ..resources import AdminResource
from ..aws import sdb_now_add


class AdminServer(BaseServer):
    
    peercheckloop = None 
    exposed_functions = []
    exposed_function_resources = {}
    
    def __init__(self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_sdb_reservation_domain, 
            aws_s3_http_cache_bucket=None,
            aws_s3_storage_bucket=None,
            aws_sdb_coordination_domain=None,
            port=5003, 
            log_file='adminserver.log',
            log_directory=None,
            log_level="debug",
            name=None,
            time_offset=None,
            peer_check_interval=60):
        if name == None:
            name = "AWSpider Admin Server UUID: %s" % self.uuid
        resource = AdminResource(self)
        self.site_port = reactor.listenTCP(port, server.Site(resource))
        self.peer_check_interval = int(peer_check_interval)
        BaseServer.__init__(
            self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_s3_http_cache_bucket=aws_s3_http_cache_bucket, 
            aws_sdb_reservation_domain=aws_sdb_reservation_domain, 
            aws_s3_storage_bucket=aws_s3_storage_bucket,
            aws_sdb_coordination_domain=aws_sdb_coordination_domain,
            log_file=log_file,
            log_directory=log_directory,
            log_level=log_level,
            name=name,
            port=port,
            time_offset=time_offset)
        
    def start(self):
        reactor.callWhenRunning(self._start)
        return self.start_deferred

    def _start(self):
        deferreds = []
        if self.time_offset is None:
            deferreds.append(self.getTimeOffset())
        d = DeferredList(deferreds, consumeErrors=True)
        d.addCallback(self._startCallback)

    def _startCallback(self, data):
        for row in data:
            if row[0] == False:
                d = self.shutdown()
                d.addCallback(self._startHandleError, row[1])
                return d
        d = BaseServer.start(self)
        d.addCallback(self._startCallback2)

    def _startCallback2(self, data):
        LOGGER.debug("At callback.")
        if self.shutdown_trigger_id is not None:
            if self.aws_sdb_coordination_domain is not None:
                self.peercheckloop = task.LoopingCall(self.peerCheck)
                self.peercheckloop.start(self.peer_check_interval)
        
    def shutdown(self):
        deferreds = []
        LOGGER.debug("%s stopping on main HTTP interface." % self.name)
        d = self.site_port.stopListening()
        if isinstance(d, Deferred):
            deferreds.append(d)
        if len(deferreds) > 0:
            d = DeferredList(deferreds)
            d.addCallback(self._shutdownCallback)
            return d
        else:
            return self._shutdownCallback(None)

    def _shutdownCallback(self, data):
        return BaseServer.shutdown(self)

    def peerCheck(self):
        sql = "SELECT * FROM `%s` WHERE created > '%s'" % (
            self.aws_sdb_coordination_domain, 
            sdb_now_add(self.peer_check_interval * -2, 
            offset=self.time_offset))
        LOGGER.debug("Querying SimpleDB, \"%s\"" % sql)
        d = self.sdb.select(sql)
        d.addCallback(self._peerCheckCallback)
        d.addErrback(self._peerCheckErrback)
        return d
    
    def _peerCheckCallback(self, data):
        LOGGER.info("Got server data:\n%s" % PRETTYPRINTER.pformat(data.values()))
        ip_addresses = [x['public_ip'][0] for x in data.values()]
        LOGGER.info("%s machines responding. %s" % (len(data), ip_addresses))
        
    def _peerCheckErrback(self, error):
        LOGGER.error("Could not query SimpleDB for peers: %s" % str(error))
    
    def clearHTTPCache(self):
        return self.s3.emptyBucket(self.aws_s3_http_cache_bucket)

    



from twisted.python.failure import Failure
from twisted.web.resource import Resource
import cStringIO, gzip
import traceback
import simplejson

class BaseResource(Resource):
    
    def __init__(self):
        Resource.__init__(self)

    def _successResponse(self, data):
        return simplejson.dumps(data)

    def _errorResponse(self, error):
        reason = str(error.value)
        tb = traceback.format_exc(traceback.extract_tb(error.tb))
        return simplejson.dumps({"error":reason, "traceback":tb})

    def _immediateResponse(self, data, request):
        encoding = request.getHeader("accept-encoding")
        if encoding and "gzip" in encoding:
            zbuf = cStringIO.StringIO()
            zfile = gzip.GzipFile(None, 'wb', 9, zbuf)
            if isinstance(data, unicode):
                zfile.write(unicode(data).encode("utf-8"))
            elif isinstance(data, str):
                zfile.write(unicode(data, 'utf-8').encode("utf-8"))
            else:
                zfile.write(unicode(data).encode("utf-8"))
            zfile.close()
            request.setHeader("Content-encoding","gzip")
            request.write(zbuf.getvalue())
        else:
            request.write(data)
        request.finish()


import sys
from twisted.web import server
from twisted.internet.defer import maybeDeferred 
from .base import BaseResource

class ExposedResource(BaseResource):
    
    isLeaf = True
    
    def __init__(self, interfaceserver, function_name):
        self.interfaceserver = interfaceserver
        self.function_name = function_name
        BaseResource.__init__(self)
    
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        kwargs = {}
        for key in request.args:
            kwargs[key] = request.args[key][0]
        d = maybeDeferred(self.interfaceserver.createReservation, self.function_name, **kwargs)
        d.addCallback(self._successResponse)
        d.addErrback(self._errorResponse)
        d.addCallback(self._immediateResponse, request)
        return server.NOT_DONE_YET



import pprint
import urllib
from uuid import uuid4
from twisted.internet.defer import Deferred, DeferredList, maybeDeferred
from twisted.internet.threads import deferToThread
from twisted.web.resource import Resource
from twisted.internet import reactor
from twisted.web import server
from .base import BaseServer, LOGGER
from ..resources import InterfaceResource, ExposedResource
from ..aws import sdb_now
from ..evaluateboolean import evaluateBoolean
from boto.ec2.connection import EC2Connection

PRETTYPRINTER = pprint.PrettyPrinter(indent=4)

class InterfaceServer(BaseServer):
    scheduler_server = None
    exposed_functions = []
    exposed_function_resources = {}
    name = "AWSpider Interface Server UUID: %s" % str(uuid4())
    
    def __init__(self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_s3_http_cache_bucket=None,
            aws_s3_storage_bucket=None,
            scheduler_server_group='flavors_spider_production',
            schedulerserver_port=5004,
            max_simultaneous_requests=50,
            max_requests_per_host_per_second=1,
            max_simultaneous_requests_per_host=5,
            port=5000, 
            log_file='interfaceserver.log',
            log_directory=None,
            log_level="debug"):
        self.aws_access_key_id=aws_access_key_id
        self.aws_secret_access_key=aws_secret_access_key
        self.scheduler_server_group=scheduler_server_group
        self.schedulerserver_port=schedulerserver_port
        resource = Resource()
        interface_resource = InterfaceResource(self)
        resource.putChild("interface", interface_resource)
        self.function_resource = Resource()
        resource.putChild("function", self.function_resource)
        self.site_port = reactor.listenTCP(port, server.Site(resource))
        BaseServer.__init__(
            self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_s3_http_cache_bucket=aws_s3_http_cache_bucket, 
            aws_s3_storage_bucket=aws_s3_storage_bucket,
            max_simultaneous_requests=max_simultaneous_requests,
            max_requests_per_host_per_second=max_requests_per_host_per_second,
            max_simultaneous_requests_per_host=max_simultaneous_requests_per_host,
            log_file=log_file,
            log_directory=log_directory,
            log_level=log_level,
            port=port)
        
    def start(self):
        deferToThread(self.setSchedulerServer)
        reactor.callWhenRunning(self._start)
        return self.start_deferred

    def _start(self):
        deferreds = []
        d = DeferredList(deferreds, consumeErrors=True)
        d.addCallback(self._startCallback)

    def _startCallback(self, data):
        for row in data:
            if row[0] == False:
                d = self.shutdown()
                d.addCallback(self._startHandleError, row[1])
                return d
        d = BaseServer.start(self)

    def shutdown(self):
        deferreds = []
        LOGGER.debug("%s stopping on main HTTP interface." % self.name)
        d = self.site_port.stopListening()
        if isinstance(d, Deferred):
            deferreds.append(d)
        if len(deferreds) > 0:
            d = DeferredList(deferreds)
            d.addCallback(self._shutdownCallback)
            return d
        else:
            return self._shutdownCallback(None)

    def _shutdownCallback(self, data):
        return BaseServer.shutdown(self)
        
    def setSchedulerServer(self):
        LOGGER.info('Locating scheduler server for security group: %s' % self.scheduler_server_group)
        if self.scheduler_server_group:
            conn = EC2Connection(self.aws_access_key_id, self.aws_secret_access_key)
            scheduler_hostnames = []
            scheduler_hostnames_a = scheduler_hostnames.append
            for reservation in conn.get_all_instances():
                for reservation_group in reservation.groups:
                    if reservation_group.id == self.scheduler_server_group:
                        for instance in reservation.instances:
                            if instance.state == "running":
                                scheduler_hostnames_a(instance.private_dns_name)
            if scheduler_hostnames:
                self.scheduler_server = scheduler_hostnames[0]
        else:
            self.scheduler_server = "0.0.0.0"
        LOGGER.debug('Scheduler Server found at %s' % self.scheduler_server)
    
    def makeCallable(self, func, interval=0, name=None, expose=False):
        function_name = BaseServer.makeCallable(
            self, 
            func, 
            interval=interval, 
            name=name, 
            expose=expose)
        if expose:
            self.exposed_functions.append(function_name)
            er = ExposedResource(self, function_name)
            function_name_parts = function_name.split("/")
            if len(function_name_parts) > 1:
                if function_name_parts[0] in self.exposed_function_resources:
                    r = self.exposed_function_resources[function_name_parts[0]]
                else:
                    r = Resource()
                    self.exposed_function_resources[function_name_parts[0]] = r
                self.function_resource.putChild(function_name_parts[0], r)
                r.putChild(function_name_parts[1], er)
            else:
                self.function_resource.putChild(function_name_parts[0], er)
            LOGGER.info("Function %s is now available via the HTTP interface." % function_name)
            
    def createReservation(self, function_name, **kwargs):
        uuid = uuid4().hex
        if not isinstance(function_name, str):
            for key in self.functions:
                if self.functions[key]["function"] == function_name:
                    function_name = key
                    break
        if function_name not in self.functions:
            raise Exception("Function %s does not exist." % function_name)
        d = self.callExposedFunction(
            self.functions[function_name]["function"], 
            kwargs, 
            function_name, 
            uuid=uuid)
        d.addCallback(self._createReservationCallback, function_name, uuid)
        d.addErrback(self._createReservationErrback, function_name, uuid)
        return d

    def _createReservationCallback(self, data, function_name, uuid):
        if self.scheduler_server:
            parameters = {
                'uuid': uuid
            }
            url = 'http://%s:%s/function/schedulerserver/remoteaddtoheap' % (self.scheduler_server, self.schedulerserver_port)
            LOGGER.info('Sending UUID to scheduler: %s' % url)
            query_string = urllib.urlencode(parameters)       
            d = self.getPage(url=url, postdata=query_string)
            d.addCallback(self._createReservationCallback2, data)
            d.addErrback(self._createReservationErrback, function_name, uuid)
            return d
        else:
            LOGGER.error('No scheduler server defined...')
            raise

    def _createReservationCallback2(self, data):
        return data

    def _createReservationErrback(self, error, function_name, uuid):
        LOGGER.error("Unable to create reservation for %s:%s, %s.\n" % (function_name, uuid, error))
        return {}


import base64
import hmac
import hashlib
import urllib
from datetime import datetime
import xml.etree.cElementTree as ET
from .lib import safe_quote_tuple, etree_to_dict
from ..requestqueuer import RequestQueuer

PA_NAMESPACE = "{http://webservices.amazon.com/AWSECommerceService/2009-10-01}"

class AmazonProductAdvertising:
    """
    Amazon Product Advertising API.
    """
    host = "ecs.amazonaws.com"
    
    def __init__(self, aws_access_key_id, aws_secret_access_key, rq=None):
        """
        **Arguments:**
         * *aws_access_key_id* -- Amazon AWS access key ID
         * *aws_secret_access_key* -- Amazon AWS secret access key
       
        **Keyword arguments:**
         * *rq* -- Optional RequestQueuer object.
        """
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
        self.rq.setHostMaxRequestsPerSecond(self.host, 0)
        self.rq.setHostMaxSimultaneousRequests(self.host, 0)
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
    
    def itemSearch(self, **kwargs):
        if "MerchantId" not in kwargs:
            kwargs["MerchantId"] = "All"
        if "SearchIndex" not in kwargs:
            kwargs["SearchIndex"] = "All"
        parameters = kwargs
        parameters["Operation"] = "ItemSearch"
        d = self._request(parameters)
        d.addCallback(self._itemSearchCallback)
        return d
        
    def _itemSearchCallback(self, data):
        xml = ET.fromstring(data["response"])
        errors = xml.findall(".//%sError" % PA_NAMESPACE)
        if len(errors) > 0:
            message = errors[0].find("./%sMessage" % PA_NAMESPACE).text
            raise Exception(message)
        items = xml.findall(".//%sItem" % PA_NAMESPACE)
        results = []
        for item in items:
            results.append(self._processItem(item))
        return results
    
    def _processItem(self, item_etree):
        item = {}
        item["asin"] = item_etree.find("./%sASIN" % PA_NAMESPACE).text
        item["detailpageurl"] = item_etree.find("./%sDetailPageURL" % PA_NAMESPACE).text
        # Attributes
        attributes = item_etree.find("./%sItemAttributes" % PA_NAMESPACE).getchildren()
        for element in attributes:
            tag = element.tag.replace(PA_NAMESPACE, "")
            item[tag.lower()] = element.text
        # Links
        links = item_etree.findall(".//%sItemLink" % PA_NAMESPACE)
        if len(links) > 0:
            link_list = []
            for link in links:
                link_list.append({
                    "description":link.find("./%sDescription" % PA_NAMESPACE).text,
                    "url":link.find("./%sURL" % PA_NAMESPACE).text})
            item["links"] = link_list
        # Sales Rank
        sales_rank = item_etree.find(".//%sSalesRank" % PA_NAMESPACE)
        if sales_rank is not None:
            item["sales_rank"] = sales_rank.text
        # Images
        item.update(self._processImageSet(item_etree))
        image_sets = item_etree.findall(".//%sImageSet" % PA_NAMESPACE)
        if len(image_sets) > 0:
            item["image_sets"] = {}
            for image_set in image_sets:
                item["image_sets"][image_set.attrib["Category"]] = self._processImageSet(image_set)
        # Subjects
        subjects = item_etree.findall(".//%sSubject" % PA_NAMESPACE)
        if len(subjects) > 0:
            item["subjects"] = []
            for subject in subjects:
                item["subjects"].append(subject.text)
        # Offer summary
        offer_summary = item_etree.find(".//%sOfferSummary" % PA_NAMESPACE)
        item["offer_summary"] = etree_to_dict(
            offer_summary, 
            namespace=PA_NAMESPACE, 
            tag_list=False, 
            convert_camelcase=True)
        # Offers 
        total_offers = item_etree.find(".//%sTotalOffers" % PA_NAMESPACE)
        if total_offers is not None:
            item["total_offers"] = total_offers.text
        total_offer_pages = item_etree.find(".//%sTotalOfferPages" % PA_NAMESPACE)
        if total_offer_pages is not None:
            item["total_offer_pages"] = total_offer_pages.text
        offers = item_etree.findall(".//%sOffer" % PA_NAMESPACE)
        if len(offers) > 0:
            item["offers"] = []
            for offer in offers:
                item["offers"].append(etree_to_dict(
                    offer, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))
        # Reviews 
        average_rating = item_etree.find(".//%sAverageRating" % PA_NAMESPACE)
        if average_rating is not None:
            item["average_rating"] = average_rating.text
        total_reviews = item_etree.find(".//%sTotalReviews" % PA_NAMESPACE)
        if total_reviews is not None:
            item["total_reviews"] = total_reviews.text
        total_review_pages = item_etree.find(".//%sTotalReviewPages" % PA_NAMESPACE)
        if total_review_pages is not None:
            item["total_review_pages"] = total_review_pages.text        
        reviews = item_etree.findall(".//%sReview" % PA_NAMESPACE)
        if len(reviews) > 0:
            item["reviews"] = []
            for review in reviews:
                item["reviews"].append(etree_to_dict(
                    review, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))
        # Editorial reviews 
        editorial_reviews = item_etree.findall(".//%sEditorialReview" % PA_NAMESPACE)
        if len(editorial_reviews) > 0:
            item["editorial_reviews"] = []
            for review in editorial_reviews:
                item["editorial_reviews"].append(etree_to_dict(
                    review, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))
        # Similar products 
        similar_products = item_etree.findall(".//%sSimilarProduct" % PA_NAMESPACE)
        if len(similar_products) > 0:
            item["similar_products"] = []
            for product in similar_products:
                item["similar_products"].append(etree_to_dict(
                    product, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))       
        # Tags
        if item_etree.find(".//%sTags" % PA_NAMESPACE) is not None:
            item["tag_information"] = {}
            distinct_tags = item_etree.find(".//%sDistinctTags" % PA_NAMESPACE)
            if distinct_tags is not None:
                item["tag_information"]["distinct_tags"] = distinct_tags.text
            distinct_items = item_etree.find(".//%sDistinctItems" % PA_NAMESPACE)
            if distinct_items is not None:
                item["tag_information"]["distinct_items"] = distinct_items.text
            distinct_users = item_etree.find(".//%sDistinctUsers" % PA_NAMESPACE)
            if distinct_users is not None:
                item["tag_information"]["distinct_users"] = distinct_users.text
            total_usages = item_etree.find(".//%sTotalUsages" % PA_NAMESPACE)
            if total_usages is not None:
                item["tag_information"]["total_usages"] = total_usages.text
            first_tagging = item_etree.find(".//%sFirstTagging" % PA_NAMESPACE)
            if first_tagging is not None:
                item["tag_information"]["first_tagging"] = etree_to_dict(
                    first_tagging, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True)
            last_tagging = item_etree.find(".//%sLastTagging" % PA_NAMESPACE)
            if last_tagging is not None:
                item["tag_information"]["last_tagging"] = etree_to_dict(
                    last_tagging, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True)
        tags = item_etree.findall(".//%sTag" % PA_NAMESPACE)
        if len(tags) > 0:
            item["tags"] = []
            for tag in tags:
                item["tags"].append(etree_to_dict(
                    tag, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))
        # BrowseNodes
        browse_nodes = item_etree.find(".//%sBrowseNodes" % PA_NAMESPACE)
        if browse_nodes is not None:
            item["browse_nodes"] = []
            for browse_node in browse_nodes.getchildren():
                item["browse_nodes"].append(etree_to_dict(
                    browse_node, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))
        # Lists
        listmania_lists = item_etree.findall(".//%sListmaniaList" % PA_NAMESPACE)
        if len(listmania_lists) > 0:
            item["listmania_lists"] = []
            for listmania_list in listmania_lists:
                item["listmania_lists"].append(etree_to_dict(
                    listmania_list, 
                    namespace=PA_NAMESPACE, 
                    tag_list=False, 
                    convert_camelcase=True))        
        return item
    
    def _processImageSet(self, item_etree):
        item = {}
        swatch_image = item_etree.find("./%sSwatchImage" % PA_NAMESPACE)
        if swatch_image is not None:
            item["swatch_image"] = swatch_image.find("./%sURL" % PA_NAMESPACE).text
        small_image = item_etree.find("./%sSmallImage" % PA_NAMESPACE)
        if small_image is not None:
            item["small_image"] = small_image.find("./%sURL" % PA_NAMESPACE).text  
        thumbnail_image = item_etree.find("./%sThumbnailImage" % PA_NAMESPACE)
        if thumbnail_image is not None:
            item["thumbnail_image"] = thumbnail_image.find("./%sURL" % PA_NAMESPACE).text
        tiny_image = item_etree.find("./%sTinyImage" % PA_NAMESPACE)
        if tiny_image is not None:
            item["tiny_image"] = tiny_image.find("./%sURL" % PA_NAMESPACE).text
        medium_image = item_etree.find("./%sMediumImage" % PA_NAMESPACE)
        if medium_image is not None:
            item["medium_image"] = medium_image.find("./%sURL" % PA_NAMESPACE).text
        large_image = item_etree.find("./%sLargeImage" % PA_NAMESPACE)
        if large_image is not None:
            item["large_image"] = large_image.find("./%sURL" % PA_NAMESPACE).text
        return item
        
    def _request(self, parameters):
        """
        Add authentication parameters and make request to Amazon.

        **Arguments:**
         * *parameters* -- Key value pairs of parameters
        """
        parameters["Service"] = "AWSECommerceService"
        parameters = self._getAuthorization("GET", parameters)
        query_string = urllib.urlencode(parameters)
        url = "https://%s/onca/xml?%s" % (self.host, query_string)
        d = self.rq.getPage(url, method="GET")
        return d

    def _canonicalize(self, parameters):
        """
        Canonicalize parameters for use with AWS Authorization.

        **Arguments:**
         * *parameters* -- Key value pairs of parameters

        **Returns:**
         * A safe-quoted string representation of the parameters.
        """
        parameters = parameters.items()
        parameters.sort(lambda x, y:cmp(x[0], y[0]))
        return "&".join([safe_quote_tuple(x) for x in parameters])

    def _getAuthorization(self, method, parameters):
        """
        Create authentication parameters.

        **Arguments:**
         * *method* -- HTTP method of the request
         * *parameters* -- Key value pairs of parameters

        **Returns:**
         * A dictionary of authorization parameters
        """
        signature_parameters = {
            "AWSAccessKeyId":self.aws_access_key_id,
            'Timestamp':datetime.utcnow().isoformat()[0:19]+"+00:00",
            "AWSAccessKeyId":self.aws_access_key_id,
            "Version":"2009-10-01"
        }
        signature_parameters.update(parameters)
        query_string = self._canonicalize(signature_parameters)
        string_to_sign = "%(method)s\n%(host)s\n%(resource)s\n%(qs)s" % {
            "method":method,
            "host":self.host.lower(),
            "resource":"/onca/xml",
            "qs":query_string,
        }
        args = [self.aws_secret_access_key, string_to_sign, hashlib.sha256]
        signature = base64.encodestring(hmac.new(*args).digest()).strip()
        signature_parameters.update({'Signature': signature})
        return signature_parameters



import inspect
import types 


__all__ = ["aliases", "expose", "make_callable", "AWSpiderPlugin"]
EXPOSED_FUNCTIONS = {}
CALLABLE_FUNCTIONS = {}
MEMOIZED_FUNCTIONS = {}
FUNCTION_ALIASES = {}

def aliases(*args):
    def decorator(f):
        FUNCTION_ALIASES[id(f)] = args
        return f
    return decorator
    
def expose(func=None, interval=0, name=None, memoize=False):
    if func is not None:
        EXPOSED_FUNCTIONS[id(func)] = {"interval":interval, "name":name}
        return func
    def decorator(f):
        EXPOSED_FUNCTIONS[id(f)] = {"interval":interval, "name":name}
        return f
    return decorator


def make_callable(func=None, interval=0, name=None, memoize=False):
    if func is not None:
        CALLABLE_FUNCTIONS[id(func)] = {"interval":interval, "name":name}
        return func
    def decorator(f):
        CALLABLE_FUNCTIONS[id(f)] = {"interval":interval, "name":name}
        return f
    return decorator


class AWSpiderPlugin(object):
    
    def __init__(self, spider):
        self.spider = spider
        check_method = lambda x:isinstance(x[1], types.MethodType)
        instance_methods = filter(check_method, inspect.getmembers(self))
        for instance_method in instance_methods:
            instance_id = id(instance_method[1].__func__)
            if instance_id in EXPOSED_FUNCTIONS:
                self.spider.expose(
                    instance_method[1],
                    interval=EXPOSED_FUNCTIONS[instance_id]["interval"],
                    name=EXPOSED_FUNCTIONS[instance_id]["name"])
                if instance_id in FUNCTION_ALIASES:
                    for name in FUNCTION_ALIASES[instance_id]:
                        self.spider.expose(
                            instance_method[1],
                            interval=EXPOSED_FUNCTIONS[instance_id]["interval"],
                            name=name)                        
            if instance_id in CALLABLE_FUNCTIONS:
                self.spider.expose(
                    instance_method[1],
                    interval=CALLABLE_FUNCTIONS[instance_id]["interval"],
                    name=CALLABLE_FUNCTIONS[instance_id]["name"])
                if instance_id in FUNCTION_ALIASES:
                    for name in CALLABLE_FUNCTIONS[instance_id]:
                        self.spider.expose(
                            instance_method[1],
                            interval=CALLABLE_FUNCTIONS[instance_id]["interval"],
                            name=name)
                            
    def setReservationFastCache(self, uuid, data):
        return self.spider.setReservationFastCache(uuid, data)
   
    def setReservationCache(self, uuid, data):
        return self.spider.setReservationCache(uuid, data)
    
    def getPage(self, *args, **kwargs):
        return self.spider.getPage(*args, **kwargs)
        